import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as data
import torch.optim as optim
import itertools
import tqdm
import copy
import scipy.stats as st
import os
import time
import cv2
from utils.vis_utils import append_text_to_image

from scipy.stats import norm
import itertools

# listwise comparison
# from models.losses import listNet

device = 'cuda:0' if torch.cuda.is_available() else 'cpu'


def gen_net(in_size=1, out_size=1, H=128, n_layers=3, activation='tanh'):
    net = []
    for i in range(n_layers):
        net.append(nn.Linear(in_size, H))
        if not activation == 'none':
            net.append(nn.LeakyReLU())
        in_size = H
    net.append(nn.Linear(in_size, out_size))
    if activation == 'tanh':
        net.append(nn.Tanh())
    elif activation == 'sig':
        net.append(nn.Sigmoid())
    elif activation == 'none':
        # linear model
        pass
    else:
        net.append(nn.ReLU())

    return net


def KCenterGreedy(obs, full_obs, num_new_sample):
    selected_index = []
    current_index = list(range(obs.shape[0]))
    new_obs = obs
    new_full_obs = full_obs
    start_time = time.time()
    for count in range(num_new_sample):
        dist = compute_smallest_dist(new_obs, new_full_obs)
        max_index = torch.argmax(dist)
        max_index = max_index.item()

        if count == 0:
            selected_index.append(max_index)
        else:
            selected_index.append(current_index[max_index])
        current_index = current_index[0:max_index] + current_index[max_index + 1:]

        new_obs = obs[current_index]
        new_full_obs = np.concatenate([full_obs, obs[selected_index]], axis=0)
    return selected_index


def compute_smallest_dist(obs, full_obs):
    obs = torch.from_numpy(obs).float()
    full_obs = torch.from_numpy(full_obs).float()
    batch_size = 100
    with torch.no_grad():
        total_dists = []
        for full_idx in range(len(obs) // batch_size + 1):
            full_start = full_idx * batch_size
            if full_start < len(obs):
                full_end = (full_idx + 1) * batch_size
                dists = []
                for idx in range(len(full_obs) // batch_size + 1):
                    start = idx * batch_size
                    if start < len(full_obs):
                        end = (idx + 1) * batch_size
                        dist = torch.norm(
                            obs[full_start:full_end, None, :].to(device) - full_obs[None, start:end, :].to(device),
                            dim=-1, p=2
                        )
                        dists.append(dist)
                dists = torch.cat(dists, dim=1)
                small_dists = torch.torch.min(dists, dim=1).values
                total_dists.append(small_dists)

        total_dists = torch.cat(total_dists)
    return total_dists.unsqueeze(1)


class RewardModel:
    def __init__(self, ds, da, 
                 ensemble_size=3, lr=3e-4, mb_size=128, size_segment=1,
                 env_maker=None, max_size=100, activation='tanh', capacity=5e5,
                 large_batch=1, label_margin=0.0,
                 teacher_beta=-1, teacher_gamma=1,
                 teacher_eps_mistake=0,
                 teacher_eps_skip=0,
                 teacher_eps_equal=0,
                 reward_model='pairwise',
                 sequential_num_samples=None,
                 alpha=0.5,optimizer='adam', schedule=False,
                 active_challenger_alpha=2e-5):

        # train data is trajectories, must process to sa and s..
        self.ds = ds
        self.da = da
        self.de = ensemble_size
        self.lr = lr
        self.ensemble = []
        self.paramlst = []
        self.opt = None
        self.model = None
        self.max_size = max_size
        self.activation = activation
        self.size_segment = size_segment
        self.alpha = alpha # disagree : entropy ratio
        self.iteration = 0
        self.schedule = schedule
        self.active_challenger_alpha = active_challenger_alpha

        self.capacity = int(capacity)
        self.buffer_seg1 = np.empty((self.capacity, size_segment, self.ds + self.da), dtype=np.float32)
        self.buffer_seg2 = np.empty((self.capacity, size_segment, self.ds + self.da), dtype=np.float32)
        self.buffer_label = np.empty((self.capacity, 1), dtype=np.float32)
        self.buffer_index = 0
        self.buffer_full = False

        self.construct_ensemble(optimizer=optimizer)
        self.inputs = []
        self.targets = []
        self.raw_actions = []
        self.img_inputs = []
        self.mb_size = mb_size
        self.origin_mb_size = mb_size
        self.train_batch_size = 128
        self.CEloss = nn.CrossEntropyLoss()
        self.running_means = []
        self.running_stds = []
        self.best_seg = []
        self.best_label = []
        self.best_action = []
        self.large_batch = large_batch

        # new teacher
        self.teacher_beta = teacher_beta
        self.teacher_gamma = teacher_gamma
        self.teacher_eps_mistake = teacher_eps_mistake
        self.teacher_eps_equal = teacher_eps_equal
        self.teacher_eps_skip = teacher_eps_skip
        self.teacher_thres_skip = 0
        self.teacher_thres_equal = 0

        self.label_margin = label_margin
        self.label_target = 1 - 2 * self.label_margin

        # compare
        self.model = reward_model
        if sequential_num_samples is None:
            self.sequential_num_samples = "plus_1"
        else:
            self.sequential_num_samples = sequential_num_samples

    def softXEnt_loss(self, input, target):
        logprobs = torch.nn.functional.log_softmax(input, dim=1)
        return -(target * logprobs).sum() / input.shape[0]

    def change_batch(self, new_frac):
        self.mb_size = int(self.origin_mb_size * new_frac)

    def set_batch(self, new_batch):
        self.mb_size = int(new_batch)

    def set_teacher_thres_skip(self, new_margin):
        self.teacher_thres_skip = new_margin * self.teacher_eps_skip

    def set_teacher_thres_equal(self, new_margin):
        self.teacher_thres_equal = new_margin * self.teacher_eps_equal

    def construct_ensemble(self, optimizer="adam"):
        for i in range(self.de):
            model = nn.Sequential(*gen_net(in_size=self.ds + self.da, out_size=1, H=256, n_layers=3,
                                           activation=self.activation)).float().to(device)
            self.ensemble.append(model)
            self.paramlst.extend(model.parameters())

        if optimizer == "adam":
            self.opt = torch.optim.Adam(self.paramlst, lr=self.lr)
        elif optimizer == "adagrad":
            self.opt = torch.optim.Adagrad(self.paramlst, lr=self.lr)

    def add_data(self, obs, act, rew, done, img_obs=None):
        sa_t = np.concatenate([obs, act], axis=-1)
        r_t = rew

        flat_input = sa_t.reshape(1, self.da + self.ds)
        r_t = np.array(r_t)
        flat_target = r_t.reshape(1, 1)

        init_data = len(self.inputs) == 0
        if init_data:
            self.inputs.append(flat_input)
            self.targets.append(flat_target)
            if img_obs is not None:
                self.img_inputs.append(img_obs)
        elif done:
            self.inputs[-1] = np.concatenate([self.inputs[-1], flat_input])
            self.targets[-1] = np.concatenate([self.targets[-1], flat_target])
            if img_obs is not None:
                self.img_inputs[-1] = np.concatenate([self.img_inputs[-1], img_obs])
            # FIFO
            if len(self.inputs) > self.max_size:
                self.inputs = self.inputs[1:]
                self.targets = self.targets[1:]
                if img_obs is not None:
                    self.img_inputs = self.img_inputs[1:]
            self.inputs.append([])
            self.targets.append([])
            if img_obs is not None:
                self.img_inputs.append([])
        else:
            if len(self.inputs[-1]) == 0:
                self.inputs[-1] = flat_input
                self.targets[-1] = flat_target
                if img_obs is not None:
                    self.img_inputs[-1] = img_obs
            else:
                self.inputs[-1] = np.concatenate([self.inputs[-1], flat_input])
                self.targets[-1] = np.concatenate([self.targets[-1], flat_target])
                if img_obs is not None:
                    self.img_inputs[-1] = np.concatenate([self.img_inputs[-1], img_obs])

    def add_data_batch(self, obses, rewards):
        num_env = obses.shape[0]
        for index in range(num_env):
            self.inputs.append(obses[index])
            self.targets.append(rewards[index])

    def get_rank_probability(self, x_1, x_2):
        # get probability x_1 > x_2
        probs = []
        for member in range(self.de):
            probs.append(self.p_hat_member(x_1, x_2, member=member).cpu().numpy())
        probs = np.array(probs)

        return np.mean(probs, axis=0), np.std(probs, axis=0)
    
    def get_batch_disagree(self, x_1, x_2):
        total_x = np.concatenate([x_1, x_2[-1:]], axis=0)
        total_r_hat = self.r_hat_member(total_x, member=-1)
        total_r_hat = total_r_hat.sum(axis=1)
        std = np.std(total_r_hat.cpu().detach().numpy())

        return std

    def get_entropy(self, x_1, x_2):
        # get probability x_1 > x_2
        probs = []
        for member in range(self.de):
            probs.append(self.p_hat_entropy(x_1, x_2, member=member).cpu().numpy())
        probs = np.array(probs)

        return np.mean(probs, axis=0), np.std(probs, axis=0)

    def p_hat_member(self, x_1, x_2, member=-1):
        # softmaxing to get the probabilities according to eqn 1
        with torch.no_grad():
            r_hat1 = self.r_hat_member(x_1, member=member)
            r_hat2 = self.r_hat_member(x_2, member=member)
            r_hat1 = r_hat1.sum(axis=1)
            r_hat2 = r_hat2.sum(axis=1)
            r_hat = torch.cat([r_hat1, r_hat2], axis=-1)

        # taking 0 index for probability x_1 > x_2
        return F.softmax(r_hat, dim=-1)[:, 0]

    def p_hat_entropy(self, x_1, x_2, member=-1):
        # softmaxing to get the probabilities according to eqn 1
        with torch.no_grad():
            r_hat1 = self.r_hat_member(x_1, member=member)
            r_hat2 = self.r_hat_member(x_2, member=member)
            r_hat1 = r_hat1.sum(axis=1)
            r_hat2 = r_hat2.sum(axis=1)
            r_hat = torch.cat([r_hat1, r_hat2], axis=-1)

        ent = F.softmax(r_hat, dim=-1) * F.log_softmax(r_hat, dim=-1)
        ent = ent.sum(axis=-1).abs()
        return ent

    def r_hat_member(self, x, member=-1):
        # the network parameterizes r hat in eqn 1 from the paper
        return self.ensemble[member](torch.from_numpy(x).float().to(device))

    def r_hat(self, x):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized
        # but I don't understand how the normalization should be happening right now :(
        r_hats = []
        for member in range(self.de):
            r_hats.append(self.r_hat_member(x, member=member).detach().cpu().numpy())
        r_hats = np.array(r_hats)

        return np.mean(r_hats)

    def r_hat_batch(self, x):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized
        # but I don't understand how the normalization should be happening right now :(
        r_hats = []
        for member in range(self.de):
            r_hats.append(self.r_hat_member(x, member=member).detach().cpu().numpy())
        r_hats = np.array(r_hats)

        return np.mean(r_hats, axis=0)

    def r_hat_batch_grad(self, x):
        # they say they average the rewards from each member of the ensemble,
        # but I think this only makes sense if the rewards are already normalized
        # but I don't understand how the normalization should be happening right now :(
        r_hats = []
        for member in range(self.de):
            r_hats.append(self.r_hat_member(x, member=member))
        r_hats = torch.stack(r_hats)  # (self.de, n_env, 1, 1)

        return torch.mean(r_hats, axis=0)  # (n_env, 1, 1)

    def save(self, model_dir, step):
        for member in range(self.de):
            torch.save(self.ensemble[member].state_dict(), '%s/reward_model_%s_%s.pt' % (model_dir, step, member))

    def load(self, model_dir, step):
        for member in range(self.de):
            self.ensemble[member].load_state_dict(torch.load('%s/reward_model_%s_%s.pt' % (model_dir, step, member), map_location=torch.device('cpu')))

    def get_train_acc(self):
        ensemble_acc = np.array([0 for _ in range(self.de)])
        max_len = self.capacity if self.buffer_full else self.buffer_index
        total_batch_index = np.random.permutation(max_len)
        batch_size = 256
        num_epochs = int(np.ceil(max_len / batch_size))

        total = 0
        for epoch in range(num_epochs):
            last_index = (epoch + 1) * batch_size
            if (epoch + 1) * batch_size > max_len:
                last_index = max_len

            sa_t_1 = self.buffer_seg1[epoch * batch_size:last_index]
            sa_t_2 = self.buffer_seg2[epoch * batch_size:last_index]
            labels = self.buffer_label[epoch * batch_size:last_index]
            labels = torch.from_numpy(labels.flatten()).long().to(device)
            total += labels.size(0)
            for member in range(self.de):
                # get logits
                r_hat1 = self.r_hat_member(sa_t_1, member=member)
                r_hat2 = self.r_hat_member(sa_t_2, member=member)
                r_hat1 = r_hat1.sum(axis=1)
                r_hat2 = r_hat2.sum(axis=1)
                r_hat = torch.cat([r_hat1, r_hat2], axis=-1)
                _, predicted = torch.max(r_hat.data, 1)
                correct = (predicted == labels).sum().item()
                ensemble_acc[member] += correct

        ensemble_acc = ensemble_acc / total
        return np.mean(ensemble_acc)

    def get_queries(self, mb_size=20, active_challenger=False, real_human=False):
        len_traj, max_len = len(self.inputs[0]), len(self.inputs)
        img_t_1, img_t_2 = None, None

        if len(self.inputs[-1]) < len_traj:
            max_len = max_len - 1

        # get train traj
        train_inputs = np.array(self.inputs[:max_len])
        train_targets = np.array(self.targets[:max_len])
        if real_human:
            train_img_inputs = np.array(self.img_inputs[:max_len]).reshape(max_len, -1, 480, 480, 3)

        if self.model == "pairwise" or self.model == "listwise":
            batch_index_2 = np.random.choice(max_len, size=mb_size, replace=True)
            batch_index_1 = np.random.choice(max_len, size=mb_size, replace=True)
        elif self.model == "sequential" or self.model == "root":
            # choose 2*mb_size random trajectories
            # num_samples = mb_size+1 # [mb_size+1, 2*mb_size]
            if "mb_size" in self.sequential_num_samples:
                ratio = float(self.sequential_num_samples.replace("plus_", "").replace("mb_size", ""))
                num_samples = mb_size + int(mb_size*ratio)
            else:
                add_value = int(self.sequential_num_samples.replace("plus_", ""))
                num_samples = mb_size + add_value
            if not active_challenger:
                batch_index_total = np.random.choice(max_len, size=num_samples, replace=True) # 2*mb_size
                num_samples_large = num_samples
            else:
                # active challenger sampling
                batch_index_total = np.random.choice(max_len, size=num_samples*self.large_batch, replace=True)
                num_samples_large = num_samples*self.large_batch
                
            batch_index_2 = batch_index_total[1:]
            batch_index_1 = batch_index_total[:-1]

        sa_t_2 = train_inputs[batch_index_2]  # Batch x T x dim of s&a
        r_t_2 = train_targets[batch_index_2]  # Batch x T x 1

        sa_t_1 = train_inputs[batch_index_1]  # Batch x T x dim of s&a
        r_t_1 = train_targets[batch_index_1]  # Batch x T x 1

        sa_t_1 = sa_t_1.reshape(-1, sa_t_1.shape[-1])  # (Batch x T) x dim of s&a
        r_t_1 = r_t_1.reshape(-1, r_t_1.shape[-1])  # (Batch x T) x 1
        sa_t_2 = sa_t_2.reshape(-1, sa_t_2.shape[-1])  # (Batch x T) x dim of s&a
        r_t_2 = r_t_2.reshape(-1, r_t_2.shape[-1])  # (Batch x T) x 1

        if real_human:
            img_t_1 = train_img_inputs[batch_index_1]
            img_t_2 = train_img_inputs[batch_index_2]
            img_t_1 = img_t_1.reshape(-1, img_t_1.shape[-3], img_t_2.shape[-2], img_t_2.shape[-1])
            img_t_2 = img_t_2.reshape(-1, img_t_2.shape[-3], img_t_2.shape[-2], img_t_2.shape[-1])

        # Generate time index 
        if self.model == "pairwise" or self.model == "listwise":
            time_index = np.array([list(range(i*len_traj,
                                                i*len_traj+self.size_segment)) for i in range(mb_size)])
            time_index_2 = time_index + np.random.choice(len_traj-self.size_segment, size=mb_size, replace=True).reshape(-1,1)
            time_index_1 = time_index + np.random.choice(len_traj-self.size_segment, size=mb_size, replace=True).reshape(-1,1)
        elif self.model == "sequential" or self.model == "root":
            time_index_total = np.array([list(range(i*len_traj,
                                                i*len_traj+self.size_segment)) for i in range(num_samples_large-1)])
            time_index_random = np.random.choice(len_traj-self.size_segment, size=num_samples_large, replace=True).reshape(-1,1)
            time_index_1 = time_index_total + time_index_random[:-1]
            time_index_2 = time_index_total + time_index_random[1:]

        sa_t_1 = np.take(sa_t_1, time_index_1, axis=0)  # Batch x size_seg x dim of s&a
        r_t_1 = np.take(r_t_1, time_index_1, axis=0)  # Batch x size_seg x 1
        sa_t_2 = np.take(sa_t_2, time_index_2, axis=0)  # Batch x size_seg x dim of s&a
        r_t_2 = np.take(r_t_2, time_index_2, axis=0)  # Batch x size_seg x 1

        if real_human:
            img_t_1 = np.take(img_t_1, time_index_1, axis=0)
            img_t_2 = np.take(img_t_2, time_index_2, axis=0)

        if not real_human:
            # augment before calculating the labels
            if self.model == "sequential":
                # additional pairs for sequential model
                sa_t = train_inputs[batch_index_total]
                r_t = train_targets[batch_index_total]
                
                sa_t_1, sa_t_2, r_t_1, r_t_2 = self.sequential_pairwise_augment(sa_t, r_t, mb_size, num_samples, num_samples_large, time_index_random, active_challenger=active_challenger)

            elif self.model == "root":
                # additional pairs for root model
                sa_t = train_inputs[batch_index_total]
                r_t = train_targets[batch_index_total]

                sa_t_1, sa_t_2, r_t_1, r_t_2 = self.root_pairwise_augment(sa_t, r_t, mb_size, num_samples, num_samples_large, time_index_random, active_challenger=active_challenger)

        if not real_human or self.model == "pairwise": 
            time_index_random = None
        return sa_t_1, sa_t_2, r_t_1, r_t_2, img_t_1, img_t_2, time_index_random

    def put_queries(self, sa_t_1, sa_t_2, labels):
        total_sample = sa_t_1.shape[0]
        next_index = self.buffer_index + total_sample
        if next_index >= self.capacity:
            self.buffer_full = True
            maximum_index = self.capacity - self.buffer_index
            np.copyto(self.buffer_seg1[self.buffer_index:self.capacity], sa_t_1[:maximum_index])
            np.copyto(self.buffer_seg2[self.buffer_index:self.capacity], sa_t_2[:maximum_index])
            np.copyto(self.buffer_label[self.buffer_index:self.capacity], labels[:maximum_index])

            remain = total_sample - (maximum_index)
            if remain > 0:
                np.copyto(self.buffer_seg1[0:remain], sa_t_1[maximum_index:])
                np.copyto(self.buffer_seg2[0:remain], sa_t_2[maximum_index:])
                np.copyto(self.buffer_label[0:remain], labels[maximum_index:])

            self.buffer_index = remain
        else:
            np.copyto(self.buffer_seg1[self.buffer_index:next_index], sa_t_1)
            np.copyto(self.buffer_seg2[self.buffer_index:next_index], sa_t_2)
            np.copyto(self.buffer_label[self.buffer_index:next_index], labels)
            self.buffer_index = next_index

    def get_label(self, sa_t_1, sa_t_2, r_t_1, r_t_2):
        sum_r_t_1 = np.sum(r_t_1, axis=1)
        sum_r_t_2 = np.sum(r_t_2, axis=1)

        # skip the query
        if self.teacher_thres_skip > 0:
            max_r_t = np.maximum(sum_r_t_1, sum_r_t_2)
            max_index = (max_r_t > self.teacher_thres_skip).reshape(-1)
            if sum(max_index) == 0:
                return None, None, None, None, []

            sa_t_1 = sa_t_1[max_index]
            sa_t_2 = sa_t_2[max_index]
            r_t_1 = r_t_1[max_index]
            r_t_2 = r_t_2[max_index]
            sum_r_t_1 = np.sum(r_t_1, axis=1)
            sum_r_t_2 = np.sum(r_t_2, axis=1)

        # equally preferable
        margin_index = (np.abs(sum_r_t_1 - sum_r_t_2) < self.teacher_thres_equal).reshape(-1)

        # perfectly rational
        seg_size = r_t_1.shape[1]
        temp_r_t_1 = r_t_1.copy()
        temp_r_t_2 = r_t_2.copy()
        for index in range(seg_size - 1):
            temp_r_t_1[:, :index + 1] *= self.teacher_gamma
            temp_r_t_2[:, :index + 1] *= self.teacher_gamma
        sum_r_t_1 = np.sum(temp_r_t_1, axis=1)
        sum_r_t_2 = np.sum(temp_r_t_2, axis=1)

        rational_labels = 1 * (sum_r_t_1 < sum_r_t_2)
        if self.teacher_beta > 0:  # Bradley-Terry rational model
            r_hat = torch.cat([torch.Tensor(sum_r_t_1), torch.Tensor(sum_r_t_2)], axis=-1)
            r_hat = r_hat * self.teacher_beta
            ent = F.softmax(r_hat, dim=-1)[:, 1]
            labels = torch.bernoulli(ent).int().numpy().reshape(-1, 1)
        else:
            labels = rational_labels

        # making a mistake
        len_labels = labels.shape[0]
        rand_num = np.random.rand(len_labels)
        noise_index = rand_num <= self.teacher_eps_mistake
        labels[noise_index] = 1 - labels[noise_index]

        # equally preferable
        labels[margin_index] = -1

        return sa_t_1, sa_t_2, r_t_1, r_t_2, labels
    
    def click_human_label(self, event, x, y, flags, param):
        # if the left mouse button was clicked before 480th pixel, assign 0
        if event == cv2.EVENT_LBUTTONDOWN:
            if x < 480:
                self.human_label = 0
            else:
                self.human_label = 1
    
    def get_real_human_label(self, sa_t_1, sa_t_2, r_t_1, r_t_2, img_t_1, img_t_2, time_index_random):
        # simulated teacher
        sum_r_t_1 = np.sum(r_t_1, axis=1)
        sum_r_t_2 = np.sum(r_t_2, axis=1)

        # perfectly rational
        seg_size = r_t_1.shape[1]
        temp_r_t_1 = r_t_1.copy()
        temp_r_t_2 = r_t_2.copy()
        for index in range(seg_size - 1):
            temp_r_t_1[:, :index + 1] *= self.teacher_gamma
            temp_r_t_2[:, :index + 1] *= self.teacher_gamma
        sum_r_t_1 = np.sum(temp_r_t_1, axis=1)
        sum_r_t_2 = np.sum(temp_r_t_2, axis=1)

        rational_labels = 1 * (sum_r_t_1 < sum_r_t_2)
        if self.teacher_beta > 0:  # Bradley-Terry rational model
            r_hat = torch.cat([torch.Tensor(sum_r_t_1), torch.Tensor(sum_r_t_2)], axis=-1)
            r_hat = r_hat * self.teacher_beta
            ent = F.softmax(r_hat, dim=-1)[:, 1]
            simulated_teacher_labels = torch.bernoulli(ent).int().numpy().reshape(-1, 1)
        else:
            simulated_teacher_labels = rational_labels

        # human 

        labels = []
        self.human_label = 0
        pseudo_r_t = []
        pseudo_r_t_1 = []
        pseudo_r_t_2 = []
        root_r_t_index = 0
        root_r_t = 0
        if self.model != "pairwise":
            pseudo_r_t = [0]

        if self.model == "root":
            root_frames = img_t_1[0]

        # real human
        # show animations of query pairs
        for i in range(len(sa_t_1)):
            # ith trajectory pair from img_t_1 and img_t_2
            if self.model == "pairwise" or self.model == "sequential":
                frames_1 = img_t_1[i]
            elif self.model == "root":
                frames_1 = root_frames
            frames_2 = img_t_2[i]
            concat_frame = np.concatenate((frames_1, np.repeat(frames_2[:1], repeats=len(frames_1), axis=0).reshape(-1, 480, 480, 3)), axis=2)
            concat_frame_later = np.concatenate((np.repeat(frames_1[-1:], repeats=len(frames_2), axis=0).reshape(-1, 480, 480, 3), frames_2), axis=2)
            concat_frame = np.concatenate((concat_frame, concat_frame_later), axis=0)
            while True:
                for j in range(len(concat_frame)):
                    if j < len(concat_frame) - 1:
                        string = 'frame: {}'.format(j%len(frames_1))
                    else:
                        answer = "left" if simulated_teacher_labels[i] == 0 else "right"
                        string = 'click better trajectory: left or right' # | answer: {}'.format(answer)
                    
                    # put index at each image
                    frame_j = concat_frame[j]
                    if self.model == "pairwise":
                        frame_j = cv2.putText(frame_j, str(2*i), (450-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                        frame_j = cv2.putText(frame_j, str(2*i+1), (450+480-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                    elif self.model == "sequential":
                        frame_j = cv2.putText(frame_j, str(i), (450-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                        frame_j = cv2.putText(frame_j, str(i+1), (450+480-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                    elif self.model == "root":
                        frame_j = cv2.putText(frame_j, "root ({})".format(root_r_t_index), (450-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                        frame_j = cv2.putText(frame_j, str(i+1), (450+480-240, 50), cv2.FONT_ITALIC, 1, (255,255,255), 2)
                    frame_j = append_text_to_image(frame_j, string, 0.5)
                    cv2.imshow('query: {}'.format(i), frame_j)
                    cv2.waitKey(50)
                    if j == len(concat_frame) - 1:
                        # let human click the better trajectory
                        cv2.setMouseCallback('query: {}'.format(i), self.click_human_label)
                        cv2.waitKey(5000)
                        COLOR = (0, 0, 255)
                        if self.human_label == 0:
                            cv2.rectangle(frame_j, (2, 2), (480, 478), COLOR, 5)
                        else:
                            cv2.rectangle(frame_j, (480, 2), (958, 478), COLOR, 5)
                        cv2.imshow('query: {}'.format(i), frame_j)
                        cv2.waitKey(500)
                cv2.destroyAllWindows()
                labels.append(self.human_label)
                if self.model == "pairwise":
                    if self.human_label == 0:
                        pseudo_r_t_1.append(0)
                        pseudo_r_t_2.append(1)
                    else:
                        pseudo_r_t_1.append(1)
                        pseudo_r_t_2.append(0)
                elif self.model == "sequential":
                    if self.human_label == 0:
                        pseudo_r_t.append(pseudo_r_t[-1] + 1)
                    else:
                        pseudo_r_t.append(pseudo_r_t[-1] - 1)
                elif self.model == "root":
                    if self.human_label == 1:
                        # update root
                        root_r_t_index = i+1
                        root_frames = copy.deepcopy(frames_2) #img_t_2[i]
                        root_r_t = root_r_t + 1
                        pseudo_r_t.append(root_r_t)
                    else:
                        # add as a child
                        pseudo_r_t.append(root_r_t-1)

                break
        labels = np.array(labels).reshape(-1, 1)
        if self.model != "pairwise":
            pseudo_r_t_1 = pseudo_r_t[:-1]
            pseudo_r_t_2 = pseudo_r_t[1:]

        print('real human labels:        {}'.format([str(x) for x in labels.squeeze()]))
        print('simulated teacher labels: {}'.format([str(x) for x in simulated_teacher_labels.squeeze()]))

        pseudo_r_t = np.array(pseudo_r_t).reshape(-1, 1)

        # augmentation
        mb_size = self.mb_size
        num_samples = len(sa_t_1) + 1 
        num_samples_large = num_samples * 1
        # augment before calculating the labels
        if self.model == "sequential":
            # additional pairs for sequential model
            sa_t = np.concatenate([sa_t_1, sa_t_2[-1:]], axis=0)            
            sa_t_1, sa_t_2, r_t_1, r_t_2 = self.sequential_pairwise_augment(sa_t, pseudo_r_t, mb_size, num_samples, num_samples_large, time_index_random, real_human=True)

        elif self.model == "root":
            # additional pairs for root model
            sa_t = np.concatenate([sa_t_1, sa_t_2[-1:]], axis=0)
            sa_t_1, sa_t_2, r_t_1, r_t_2 = self.root_pairwise_augment(sa_t, pseudo_r_t, mb_size, num_samples, num_samples_large, time_index_random, real_human=True)

        if self.model == "sequential" or self.model == "root":
            labels = 1 * (r_t_1 < r_t_2)
        return sa_t_1, sa_t_2, labels

    def kcenter_sampling(self):
        # get queries
        num_init = self.mb_size * self.large_batch
        sa_t_1, sa_t_2, r_t_1, r_t_2 = self.get_queries(mb_size=num_init)

        # get final queries based on kmeans clustering
        temp_sa_t_1 = sa_t_1[:, :, :self.ds]
        temp_sa_t_2 = sa_t_2[:, :, :self.ds]
        temp_sa = np.concatenate([temp_sa_t_1.reshape(num_init, -1), temp_sa_t_2.reshape(num_init, -1)], axis=1)

        max_len = self.capacity if self.buffer_full else self.buffer_index

        tot_sa_1 = self.buffer_seg1[:max_len, :, :self.ds]
        tot_sa_2 = self.buffer_seg2[:max_len, :, :self.ds]
        tot_sa = np.concatenate([tot_sa_1.reshape(max_len, -1), tot_sa_2.reshape(max_len, -1)], axis=1)

        selected_index = KCenterGreedy(temp_sa, tot_sa, self.mb_size)

        r_t_1, sa_t_1 = r_t_1[selected_index], sa_t_1[selected_index]
        r_t_2, sa_t_2 = r_t_2[selected_index], sa_t_2[selected_index]

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)

    def kcenter_disagree_sampling(self):
        num_init = self.mb_size * self.large_batch
        num_init_half = int(num_init * 0.5)

        # get queries
        sa_t_1, sa_t_2, r_t_1, r_t_2 = self.get_queries(mb_size=num_init)

        # get final queries based on uncertainty
        _, disagree = self.get_rank_probability(sa_t_1, sa_t_2)
        top_k_index = (-disagree).argsort()[:num_init_half]
        r_t_1, sa_t_1 = r_t_1[top_k_index], sa_t_1[top_k_index]
        r_t_2, sa_t_2 = r_t_2[top_k_index], sa_t_2[top_k_index]

        # get final queries based on kmeans clustering
        temp_sa_t_1 = sa_t_1[:, :, :self.ds]
        temp_sa_t_2 = sa_t_2[:, :, :self.ds]

        temp_sa = np.concatenate([temp_sa_t_1.reshape(num_init_half, -1), temp_sa_t_2.reshape(num_init_half, -1)],
                                 axis=1)

        max_len = self.capacity if self.buffer_full else self.buffer_index

        tot_sa_1 = self.buffer_seg1[:max_len, :, :self.ds]
        tot_sa_2 = self.buffer_seg2[:max_len, :, :self.ds]
        tot_sa = np.concatenate([tot_sa_1.reshape(max_len, -1), tot_sa_2.reshape(max_len, -1)], axis=1)

        selected_index = KCenterGreedy(temp_sa, tot_sa, self.mb_size)

        r_t_1, sa_t_1 = r_t_1[selected_index], sa_t_1[selected_index]
        r_t_2, sa_t_2 = r_t_2[selected_index], sa_t_2[selected_index]

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)

    def kcenter_entropy_sampling(self):
        num_init = self.mb_size * self.large_batch
        num_init_half = int(num_init * 0.5)

        # get queries
        sa_t_1, sa_t_2, r_t_1, r_t_2 = self.get_queries(mb_size=num_init)

        # get final queries based on uncertainty
        entropy, _ = self.get_entropy(sa_t_1, sa_t_2)
        top_k_index = (-entropy).argsort()[:num_init_half]
        r_t_1, sa_t_1 = r_t_1[top_k_index], sa_t_1[top_k_index]
        r_t_2, sa_t_2 = r_t_2[top_k_index], sa_t_2[top_k_index]

        # get final queries based on kmeans clustering
        temp_sa_t_1 = sa_t_1[:, :, :self.ds]
        temp_sa_t_2 = sa_t_2[:, :, :self.ds]

        temp_sa = np.concatenate([temp_sa_t_1.reshape(num_init_half, -1), temp_sa_t_2.reshape(num_init_half, -1)],
                                 axis=1)

        max_len = self.capacity if self.buffer_full else self.buffer_index

        tot_sa_1 = self.buffer_seg1[:max_len, :, :self.ds]
        tot_sa_2 = self.buffer_seg2[:max_len, :, :self.ds]
        tot_sa = np.concatenate([tot_sa_1.reshape(max_len, -1), tot_sa_2.reshape(max_len, -1)], axis=1)

        selected_index = KCenterGreedy(temp_sa, tot_sa, self.mb_size)

        r_t_1, sa_t_1 = r_t_1[selected_index], sa_t_1[selected_index]
        r_t_2, sa_t_2 = r_t_2[selected_index], sa_t_2[selected_index]

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)

    def uniform_sampling(self, real_human=False):
        # get queries
        sa_t_1, sa_t_2, r_t_1, r_t_2, img_t_1, img_t_2, time_index_random = self.get_queries(mb_size=self.mb_size, real_human=real_human)

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        # get real human labels and augment the buffer
        if real_human:
            sa_t_1, sa_t_2, labels = self.get_real_human_label(sa_t_1, sa_t_2, r_t_1, r_t_2, img_t_1, img_t_2, time_index_random)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)

    def disagreement_sampling(self):
        # get queries
        sa_t_1, sa_t_2, r_t_1, r_t_2 = self.get_queries(mb_size=self.mb_size * self.large_batch)

        # get final queries based on uncertainty
        _, disagree = self.get_rank_probability(sa_t_1, sa_t_2)
        top_k_index = (-disagree).argsort()[:self.mb_size]
        r_t_1, sa_t_1 = r_t_1[top_k_index], sa_t_1[top_k_index]
        r_t_2, sa_t_2 = r_t_2[top_k_index], sa_t_2[top_k_index]

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)

    def entropy_sampling(self):
        # get queries
        sa_t_1, sa_t_2, r_t_1, r_t_2 = self.get_queries(mb_size=self.mb_size * self.large_batch)

        # get final queries based on uncertainty
        entropy, _ = self.get_entropy(sa_t_1, sa_t_2)

        top_k_index = (-entropy).argsort()[:self.mb_size]
        r_t_1, sa_t_1 = r_t_1[top_k_index], sa_t_1[top_k_index]
        r_t_2, sa_t_2 = r_t_2[top_k_index], sa_t_2[top_k_index]

        # get labels
        sa_t_1, sa_t_2, r_t_1, r_t_2, labels = self.get_label(sa_t_1, sa_t_2, r_t_1, r_t_2)

        if len(labels) > 0:
            self.put_queries(sa_t_1, sa_t_2, labels)

        return len(labels)
    
    
    def sequential_pairwise_augment(self, sa_t, r_t, mb_size, num_samples, num_samples_large, time_index_random, active_challenger=False, real_human=False):
        add_indices = []
        sa_t_1 = sa_t[:-1]
        sa_t_2 = sa_t[1:]
        r_t_1 = r_t[:-1]
        r_t_2 = r_t[1:]

        len_traj, max_len = len(self.inputs[0]), len(self.inputs)
        if not real_human:
            # sa_t_1 and sa_t_2 time index
            
            time_index_total = np.array([list(range(i*len_traj,
                                                    i*len_traj+self.size_segment)) for i in range(num_samples_large-1)])
            time_index_1 = time_index_total + time_index_random[:-1]
            time_index_2 = time_index_total + time_index_random[1:]

            sa_t_1 = sa_t_1.reshape(-1, sa_t_1.shape[-1])  # (Batch x T) x dim of s&a
            r_t_1 = r_t_1.reshape(-1, r_t_1.shape[-1])  # (Batch x T) x 1
            sa_t_2 = sa_t_2.reshape(-1, sa_t_2.shape[-1])  # (Batch x T) x dim of s&a
            r_t_2 = r_t_2.reshape(-1, r_t_2.shape[-1])  # (Batch x T) x 1
            
            sa_t_1 = np.take(sa_t_1, time_index_1, axis=0)  # Batch x size_seg x dim of s&a
            r_t_1 = np.take(r_t_1, time_index_1, axis=0)  # Batch x size_seg x 1
            sa_t_2 = np.take(sa_t_2, time_index_2, axis=0)  # Batch x size_seg x dim of s&a
            r_t_2 = np.take(r_t_2, time_index_2, axis=0)  # Batch x size_seg x 1
        
        
        # sequential pairs augmentation
        last_index = 0
        increase_flag = (np.sum(r_t[0]) <= np.sum(r_t[1]))
        counts = []
        count = 1

        for sample_idx in range(num_samples-1):
            if increase_flag == (np.sum(r_t[sample_idx]) <= np.sum(r_t[sample_idx+1])):
                count += 1
            else:
                nC2 = list(itertools.combinations(range(last_index, sample_idx+1), 2))
                add_indices += [list(x) for x in nC2 if abs(x[0]-x[1])!= 1]

                counts.append(count)
                last_index = sample_idx
                increase_flag = not increase_flag
                count = 1
        
        if len(add_indices) > 0:
            print("add indices: ", len(add_indices))
            add_indices = np.array(add_indices)
            add_sa_t_1 = sa_t[add_indices[:, 0]]
            add_sa_t_2 = sa_t[add_indices[:, 1]]
            add_r_t_1 = r_t[add_indices[:, 0]]
            add_r_t_2 = r_t[add_indices[:, 1]]
            add_time_index_total = np.array([list(range(i*len_traj,
                                                i*len_traj+self.size_segment)) for i in range(len(add_indices))])
            if not active_challenger:
                if not real_human:
                    add_time_index_1 = add_time_index_total + time_index_random[add_indices[:, 0]]
                    add_time_index_2 = add_time_index_total + time_index_random[add_indices[:, 1]]

                    add_sa_t_1 = add_sa_t_1.reshape(-1, add_sa_t_1.shape[-1])  # (Batch x T) x dim of s&a
                    add_r_t_1 = add_r_t_1.reshape(-1, add_r_t_1.shape[-1])  # (Batch x T) x 1
                    add_sa_t_2 = add_sa_t_2.reshape(-1, add_sa_t_2.shape[-1])  # (Batch x T) x dim of s&a
                    add_r_t_2 = add_r_t_2.reshape(-1, add_r_t_2.shape[-1])  # (Batch x T) x 1

                    add_sa_t_1 = np.take(add_sa_t_1, add_time_index_1, axis=0)  # Batch x size_seg x dim of s&a
                    add_r_t_1 = np.take(add_r_t_1, add_time_index_1, axis=0)  # Batch x size_seg x 1
                    add_sa_t_2 = np.take(add_sa_t_2, add_time_index_2, axis=0)  # Batch x size_seg x dim of s&a
                    add_r_t_2 = np.take(add_r_t_2, add_time_index_2, axis=0)  # Batch x size_seg x 1

            sa_t_1 = np.concatenate((sa_t_1, add_sa_t_1), axis=0)
            sa_t_2 = np.concatenate((sa_t_2, add_sa_t_2), axis=0)
            r_t_1 = np.concatenate((r_t_1, add_r_t_1), axis=0)
            r_t_2 = np.concatenate((r_t_2, add_r_t_2), axis=0)
        
        return sa_t_1, sa_t_2, r_t_1, r_t_2

    def root_pairwise_augment(self, sa_t, r_t, mb_size, num_samples, num_samples_large, time_index_random, active_challenger=False, real_human=False):
        add_indices = []
        sa_t_1 = sa_t[:-1]
        sa_t_2 = sa_t[1:]
        r_t_1 = r_t[:-1]
        r_t_2 = r_t[1:]

        len_traj, max_len = len(self.inputs[0]), len(self.inputs)
        if not real_human:
            # sa_t_1 and sa_t_2 time index
            time_index_total = np.array([list(range(i*len_traj,
                                                    i*len_traj+self.size_segment)) for i in range(num_samples_large-1)])
            time_index_1 = time_index_total + time_index_random[:-1]
            time_index_2 = time_index_total + time_index_random[1:]

            sa_t_1 = sa_t_1.reshape(-1, sa_t_1.shape[-1])  # (Batch x T) x dim of s&a
            r_t_1 = r_t_1.reshape(-1, r_t_1.shape[-1])  # (Batch x T) x 1
            sa_t_2 = sa_t_2.reshape(-1, sa_t_2.shape[-1])  # (Batch x T) x dim of s&a
            r_t_2 = r_t_2.reshape(-1, r_t_2.shape[-1])  # (Batch x T) x 1

            sa_t_1 = np.take(sa_t_1, time_index_1, axis=0)  # Batch x size_seg x dim of s&a
            r_t_1 = np.take(r_t_1, time_index_1, axis=0)  # Batch x size_seg x 1
            sa_t_2 = np.take(sa_t_2, time_index_2, axis=0)  # Batch x size_seg x dim of s&a
            r_t_2 = np.take(r_t_2, time_index_2, axis=0)  # Batch x size_seg x 1
        
        # root pairs augmentation
        tree = {"root": [0], "children": {0:[]}}
        best_sample_idx = 0
        already_included_pairs = []
        
        
        for sample_idx in range(1, num_samples):
            if np.sum(r_t[sample_idx]) >= np.sum(r_t[best_sample_idx]):
                tree["root"].append(sample_idx)
                tree["children"][sample_idx] = [best_sample_idx]
                already_included_pairs.append([sample_idx, best_sample_idx])
                best_sample_idx = sample_idx
            else:
                tree["children"][best_sample_idx].append(sample_idx)
                already_included_pairs.append([best_sample_idx, sample_idx])
        children = tree["children"][0].copy()
        add_pairs = [[0, child] for child in children]

        if len(add_pairs)>0:
            add_indices.extend(add_pairs)
        for parent in tree["root"][1:]:
            children.extend(tree["children"][parent])
            add_pairs = [[parent, child] for child in children]
            add_indices.extend(add_pairs)

        # exclude pairs already in the batch
        add_indices = [pair for pair in add_indices if pair not in already_included_pairs]

        if len(add_indices) > 0:
            add_indices = np.array(add_indices)
            add_sa_t_1 = sa_t[add_indices[:, 0]]
            add_sa_t_2 = sa_t[add_indices[:, 1]]
            add_r_t_1 = r_t[add_indices[:, 0]]
            add_r_t_2 = r_t[add_indices[:, 1]]

            if not active_challenger:
                if not real_human:
                    add_time_index_total = np.array([list(range(i*len_traj,
                                                        i*len_traj+self.size_segment)) for i in range(len(add_indices))])
                    add_time_index_1 = add_time_index_total + time_index_random[add_indices[:, 0]]
                    add_time_index_2 = add_time_index_total + time_index_random[add_indices[:, 1]]
                    add_sa_t_1 = add_sa_t_1.reshape(-1, add_sa_t_1.shape[-1])  # (Batch x T) x dim of s&a
                    add_r_t_1 = add_r_t_1.reshape(-1, add_r_t_1.shape[-1])  # (Batch x T) x 1
                    add_sa_t_2 = add_sa_t_2.reshape(-1, add_sa_t_2.shape[-1])  # (Batch x T) x dim of s&a
                    add_r_t_2 = add_r_t_2.reshape(-1, add_r_t_2.shape[-1])  # (Batch x T) x 1

                    add_sa_t_1 = np.take(add_sa_t_1, add_time_index_1, axis=0)  # Batch x size_seg x dim of s&a
                    add_r_t_1 = np.take(add_r_t_1, add_time_index_1, axis=0)  # Batch x size_seg x 1
                    add_sa_t_2 = np.take(add_sa_t_2, add_time_index_2, axis=0)  # Batch x size_seg x dim of s&a
                    add_r_t_2 = np.take(add_r_t_2, add_time_index_2, axis=0)  # Batch x size_seg x 1

            sa_t_1 = np.concatenate((sa_t_1, add_sa_t_1), axis=0)
            sa_t_2 = np.concatenate((sa_t_2, add_sa_t_2), axis=0)
            r_t_1 = np.concatenate((r_t_1, add_r_t_1), axis=0)
            r_t_2 = np.concatenate((r_t_2, add_r_t_2), axis=0)
        
        return sa_t_1, sa_t_2, r_t_1, r_t_2
    
    def train_reward(self):
        ensemble_losses = [[] for _ in range(self.de)]
        ensemble_acc = np.array([0 for _ in range(self.de)])

        max_len = self.capacity if self.buffer_full else self.buffer_index
        total_batch_index = []
        for _ in range(self.de):
            total_batch_index.append(np.random.permutation(max_len))

        num_epochs = int(np.ceil(max_len / self.train_batch_size))
        list_debug_loss1, list_debug_loss2 = [], []
        total = 0

        for epoch in range(num_epochs):
            self.opt.zero_grad()
            loss = 0.0

            last_index = (epoch + 1) * self.train_batch_size
            if last_index > max_len:
                last_index = max_len

            for member in range(self.de):
                # get random batch
                idxs = total_batch_index[member][epoch * self.train_batch_size:last_index]
                sa_t_1 = self.buffer_seg1[idxs]
                sa_t_2 = self.buffer_seg2[idxs]
                labels = self.buffer_label[idxs]
                labels = torch.from_numpy(labels.flatten()).long().to(device)

                if member == 0:
                    total += labels.size(0)

                # get logits
                r_hat1 = self.r_hat_member(sa_t_1, member=member)
                r_hat2 = self.r_hat_member(sa_t_2, member=member)
                r_hat1 = r_hat1.sum(axis=1)
                r_hat2 = r_hat2.sum(axis=1)
                r_hat = torch.cat([r_hat1, r_hat2], axis=-1)

                # compute loss
                curr_loss = self.CEloss(r_hat, labels)
                loss += curr_loss
                ensemble_losses[member].append(curr_loss.item())

                # compute acc
                _, predicted = torch.max(r_hat.data, 1)
                correct = (predicted == labels).sum().item()
                ensemble_acc[member] += correct

            loss.backward()
            self.opt.step()

        ensemble_acc = ensemble_acc / total

        return ensemble_acc

    def train_soft_reward(self):
        ensemble_losses = [[] for _ in range(self.de)]
        ensemble_acc = np.array([0 for _ in range(self.de)])

        max_len = self.capacity if self.buffer_full else self.buffer_index
        total_batch_index = []
        for _ in range(self.de):
            total_batch_index.append(np.random.permutation(max_len))

        num_epochs = int(np.ceil(max_len / self.train_batch_size))
        list_debug_loss1, list_debug_loss2 = [], []
        total = 0
        
        for epoch in range(num_epochs):
            if self.schedule:
                self.iteration += 1
                # scale the decay of the exponential moving average scales by 1-1/self.iteration
                # scaling the learning rate by 1/np.sqrt(self.iteration)
                for group in self.opt.param_groups:
                    group['lr'] = self.lr*((self.iteration)**(-0.5))
                    group['weight_decay'] = (1-1/self.iteration)
            self.opt.zero_grad()
            loss = 0.0

            last_index = (epoch + 1) * self.train_batch_size
            if last_index > max_len:
                last_index = max_len

            for member in range(self.de):
                if self.model == "pairwise":
                    # get random batch
                    idxs = total_batch_index[member][epoch*self.train_batch_size:last_index]
                    sa_t_1 = self.buffer_seg1[idxs]
                    sa_t_2 = self.buffer_seg2[idxs]
                    labels = self.buffer_label[idxs]
                    labels = torch.from_numpy(labels.flatten()).long().to(device)
                    
                    if member == 0:
                        total += labels.size(0)
                    
                    # get logits
                    r_hat1 = self.r_hat_member(sa_t_1, member=member)
                    r_hat2 = self.r_hat_member(sa_t_2, member=member)
                    r_hat1 = r_hat1.sum(axis=1)
                    r_hat2 = r_hat2.sum(axis=1)
                    r_hat = torch.cat([r_hat1, r_hat2], axis=-1)

                    # compute loss
                    uniform_index = labels == -1
                    labels[uniform_index] = 0
                    target_onehot = torch.zeros_like(r_hat).scatter(1, labels.unsqueeze(1), self.label_target)
                    target_onehot += self.label_margin
                    if sum(uniform_index) > 0:
                        target_onehot[uniform_index] = 0.5
                    curr_loss = self.softXEnt_loss(r_hat, target_onehot)
                    loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    
                    # compute acc
                    _, predicted = torch.max(r_hat.data, 1)
                    correct = (predicted == labels).sum().item()
                    ensemble_acc[member] += correct
                elif self.model == "sequential" or self.model == "root":
                    # get random batch
                    idxs = total_batch_index[member][epoch*self.train_batch_size:last_index]
                    sa_t_1 = self.buffer_seg1[idxs]
                    sa_t_2 = self.buffer_seg2[idxs]
                    labels = self.buffer_label[idxs]
                    labels = torch.from_numpy(labels.flatten()).long().to(device)
                    
                    if member == 0:
                        total += labels.size(0)
                    
                    # get logits
                    r_hat1 = self.r_hat_member(sa_t_1, member=member)
                    r_hat2 = self.r_hat_member(sa_t_2, member=member)
                    r_hat1 = r_hat1.sum(axis=1)
                    r_hat2 = r_hat2.sum(axis=1)
                    r_hat = torch.cat([r_hat1, r_hat2], axis=-1)

                    # compute loss
                    uniform_index = labels == -1
                    labels[uniform_index] = 0
                    target_onehot = torch.zeros_like(r_hat).scatter(1, labels.unsqueeze(1), self.label_target)
                    target_onehot += self.label_margin
                    if sum(uniform_index) > 0:
                        target_onehot[uniform_index] = 0.5
                    curr_loss = self.softXEnt_loss(r_hat, target_onehot)
                    loss += curr_loss
                    ensemble_losses[member].append(curr_loss.item())
                    
                    # compute acc
                    _, predicted = torch.max(r_hat.data, 1)
                    correct = (predicted == labels).sum().item()
                    ensemble_acc[member] += correct

            loss.backward()
            self.opt.step()

        ensemble_acc = ensemble_acc / total

        return ensemble_acc
